@@ -1,14 +1,13 @@
 # microstrategy.R
 # Interface for connecting to the MicroStrategy REST API server, creating datasets, updating datasets, and extracting data from reports and cubes
 
+
 #' @import httr
 #' @importFrom jsonlite toJSON
 #' @importFrom openssl base64_encode
 #' @importFrom utils tail
 #' @importFrom methods new
 
-# TODO: Add validity check to connection object before making requests and optionally extend session token
-
 #' @title Connection class
 #'
 #' @description Base S4 class object containing connection parameters
@@ -16,8 +15,16 @@
 #' @slot password Password
 #' @slot base_url URL for the REST API server
 #' @slot project_name Name of the project to connect to (e.g. "MicroStrategy Tutorial")
-#' @slot project_id Project ID corresponding to the chosen project name. This is determined when connecting to the project by name.
+#' @slot project_id Project ID corresponding to the chosen project name. This is determined when 
+#' connecting to the project by name.
+#' @slot application_code Code used to identify the client with MicroStrategy.
+#' @slot web_version web version.
+#' @slot iserver_version iServer version.
+#' @slot VRCH Current minimum version supported.
 #' @slot login_mode Authentication option. Standard (1) or LDAP (16).
+#' @slot web_version web version.
+#' @slot iserver_version iServer version.
+#' @slot version_ok Both iServer and web version are supported.
 #' @slot ssl_verify Default TRUE. Attempts to verify SSL certificates with each request.
 #' @slot auth_token Token provided by the I-Server after a successful log in.
 #' @slot cookies Cookies returned by the I-Server after a successful log in.
@@ -30,10 +37,15 @@
                                   project_name = 'character',
                                   project_id = 'character',
                                   login_mode = 'numeric',
+                                  application_code = 'numeric',
+                                  web_version = 'character',
+                                  iserver_version = 'character',
+                                  version_ok = 'logical',
+                                  VRCH = "character",
                                   ssl_verify = 'logical',
                                   auth_token = 'character',
-                                  cookies = 'character'))
-
+                                  cookies = 'character'),
+                        prototype = list(application_code = 65, VRCH = "11.1.0400"))
 
 #' @title Create a MicroStrategy REST API connection
 #'
@@ -42,8 +54,9 @@
 #' @param username Username
 #' @param password Password
 #' @param project_name Name of the project you intend to connect to. Case-sensitive
+#' @param project_id ID of the project you intend to connect to
 #' @param login_mode Specifies the authentication mode to use. Supported authentication modes
-#'                   are: Standard (1) (default) or LDAP (16)
+#'                   are Standard (1) (default) or LDAP (16)
 #' @param ssl_verify If \code{TRUE} (default), verifies the server's SSL certificates with each request
 #' @return A connection object to use in subsequent requests
 #' @name connect_mstr
@@ -61,39 +74,61 @@
 #' close(con)
 #' }
 #' @export
-connect_mstr <- function(base_url, username, password, project_name, login_mode=1, ssl_verify=TRUE){
+connect_mstr <- function(base_url, username, password, project_name = NULL, project_id = NULL, login_mode = 1, ssl_verify = TRUE) {
 
   # Basic error checking for input types
-  if(class(base_url) != "character") stop("'base_url' must be a character; try class(base_url)")
-  if(class(username) != 'character') stop("'username' must be a character; try class(username)")
-  if(class(password) != 'character') stop("'password' must be a character; try class(password)")
-  if(class(project_name) != 'character') stop("'project_name' must be a character; try class(project_name)")
-  if(!(login_mode %in% c(1, 8, 16))) stop("Invalid login mode. Only '1' (normal), '8' (guest), or '16' (LDAP) are supported.")
-  if(class(ssl_verify) != 'logical') stop("'ssl_verify' must be TRUE or FALSE")
+  if (class(base_url) != "character") stop("'base_url' must be a character; try class(base_url)")
+  if (class(username) != 'character') stop("'username' must be a character; try class(username)")
+  if (class(password) != 'character') stop("'password' must be a character; try class(password)")
+  if (is.null(project_id) && is.null(project_name)) stop("Specify 'project_name' or 'project_id'.")
+  if (!(login_mode %in% c(1, 8, 16))) stop("Invalid login mode. Only '1' (normal), '8' (guest), or '16' (LDAP) are supported.")
+  if (class(ssl_verify) != 'logical') stop("'ssl_verify' must be TRUE or FALSE")
 
   # Creates a new connection object
-  con <- .connection(base_url=base_url, username=username, password=password,
-                     project_name=project_name, login_mode=login_mode)
-
-  if(!ssl_verify){
-    httr::set_config(config(ssl_verifypeer = FALSE))
-    con@ssl_verify <- FALSE
-  } else {
-    con@ssl_verify <- ssl_verify
+  if (!is.null(project_id)) {
+    if (class(project_id) != 'character') stop("'project_id' must be a character; try class(project_id)")
+    con <- .connection(base_url = base_url, username = username, password = password,
+                       project_id = project_id, login_mode = login_mode)
   }
+  else {
+    if (class(project_name) != 'character') stop("'project_name' must be a character; try class(project_name)")
+    con <- .connection(base_url = base_url, username = username, password = password,
+                       project_name = project_name, login_mode = login_mode)
+  }
+  # Check if iServer and Web version are supported by MSTRIO
+  info <- check_version(con@base_url, con@VRCH)
+  con@version_ok <- info$is_ok
+  con@web_version <- info$web_version
+  con@iserver_version <- info$iserver_version
 
-  # Makes connection
-  tmp_con <- connect(connection=con)
+  if (con@version_ok) {
+    if (!ssl_verify) {
+      httr::set_config(config(ssl_verifypeer = FALSE))
+      con@ssl_verify <- FALSE
+    } else {
+      con@ssl_verify <- ssl_verify
+    }
 
-  # Add authentication token and cookies to connection object
-  con@auth_token <- tmp_con$auth_token
-  con@cookies <- tmp_con$cookies
+    # Makes connection
+    tmp_con <- connect(connection = con)
 
-  # Connect to the project and set object's project id property
-  con@project_id <- select_project(connection=con)
+    # Add authentication token and cookies to connection object
+    con@auth_token <- tmp_con$auth_token
+    con@cookies <- tmp_con$cookies
 
-  # Return connection object
-  con
+    if (is.null(project_id)) {
+      # Connect to the project and set object's project id property
+      con@project_id <- select_project(connection = con)
+    }
+
+    # Return connection object
+    return(con)
+  }
+  else {
+    stop(sprintf("This version of mstrio is only supported on MicroStrategy %s or higher.
+    Current Intelligence Server version: %s
+    Current MicroStrategy Web version: %s", con@VRCH, con@web_version, con@iserver_version), call. = FALSE)
+  }
 }
 
 # TODO: Document internal-only (non-exported) function and method
@@ -105,18 +140,20 @@ setMethod("connect", "connection", function(connection){
   # Create session
   response <- login(connection = connection)
 
-  if(response$status_code != 204){
-    # Raise server error message
+  # Raises server error message
+  if(http_error(response)) {   # http != 204
     status <- http_status(response)
     errors <- content(response)
-    errmsg <- paste0("Authentication error. Check user credentials or REST API URL and try again.\n",
-                     "HTTP ", response$status_code, " ", status$reason, ", ", status$message, "\n",
-                     "I-Server Error ", errors$code, ", ", errors$message)
-    stop(errmsg)
+    usrmsg <- "Authentication error. Check user credentials or REST API URL and try again."
 
+    stop(sprintf("%s\n HTTP Error: %s %s %s\n I-Server Error: %s %s",
+                 usrmsg, response$status_code, status$reason, status$message, errors$code, errors$message),
+         call.=FALSE)
   }
+
   return(list(auth_token = as.character(response$headers['x-mstr-authtoken']),
               cookies = as.character(response$headers['set-cookie'])))
+
 })
 
 
@@ -147,15 +184,18 @@ setMethod("close", "connection", function(connection){
 
   # Terminate the connection
   response <- logout(connection = connection)
-  if(response$status_code != 204){
-    # Raise server error message
+
+  # Raises server error message
+  if(http_error(response)) {   # http != 204
     status <- http_status(response)
     errors <- content(response)
-    errmsg <- paste0("Error attempting to terminate session connection. The session may have been terminated by the Intelligence Server.\n",
-                     "HTTP ", response$status_code, " ", status$reason, ", ", status$message, "\n",
-                     "I-Server Error ", errors$code, ", ", errors$message)
-    stop(errmsg)
+    usrmsg <- "Error attempting to terminate session connection. The session may have been terminated by the Intelligence Server."
+
+    stop(sprintf("%s\n HTTP Error: %s %s %s\n I-Server Error: %s %s",
+                 usrmsg, response$status_code, status$reason, status$message, errors$code, errors$message),
+         call.=FALSE)
   }
+
 })
 
 
@@ -167,14 +207,15 @@ setMethod("select_project", "connection", function(connection){
 
   response <- projects(connection=connection)
 
-  if(response$status_code != 200){
-    # Raise server error message
+  # Raises server error message
+  if(http_error(response)) {   # http != 200
     status <- http_status(response)
     errors <- content(response)
-    errmsg <- paste0("Error connecting to project '", connection@project_name, "'. Check project name and try again.\n",
-                     "HTTP ", response$status_code, " ", status$reason, ", ", status$message, "\n",
-                     "I-Server Error ", errors$code, ", ", errors$message)
-    stop(errmsg)
+    usrmsg <- paste("Error connecting to project", connection@project_name, ". Check project name and try again.")
+
+    stop(sprintf("%s\n HTTP Error: %s %s %s\n I-Server Error: %s %s",
+                 usrmsg, response$status_code, status$reason, status$message, errors$code, errors$message),
+         call.=FALSE)
   }
 
   projs <- content(response)
@@ -184,22 +225,27 @@ setMethod("select_project", "connection", function(connection){
     }
   }
 
-  # Project not found
+  # If executing the below, it means the project was not found in the result set. Possible typo in project name parameter.
   # Close the session. Assuming the user will attempt to re-authenticate, this prevents excess sessions
   on.exit(close(connection=connection))
+
+  # Raises server error message
   status <- http_status(response)
   errors <- content(response)
-  errmsg <- paste0("Project '", connection@project_name, "' not found. Check project name and try again.\n",
-                   "HTTP ", response$status_code, " ", status$reason, ", ", status$message, "\n",
-                   "I-Server Error ", errors$code, ", ", errors$message)
-  stop(errmsg)
+  usrmsg <- paste0("Project '", connection@project_name, "' not found. Check project name and try again.")
+
+  stop(sprintf("%s\n HTTP Error: %s %s %s\n I-Server Error: %s %s",
+               usrmsg, response$status_code, status$reason, status$message, errors$code, errors$message),
+       call.=FALSE)
+
 })
 
 
 
-#' @title Extracts the contents of a report into a R Data.Frame
+#' @title Extracts the contents of a report into a R Data.Frame (deprecated)
 #'
-#' @description Extracts the contents of a MicroStrategy report into a R Data.Frame
+#' @description Extracts the contents of a MicroStrategy report into a R Data.Frame. This function is deprecated.
+#'  Check out the to_dataframe() from the Report class. \code{\link{Report}}
 #' @param connection MicroStrategy REST API connection object
 #' @param report_id Unique ID of the report you wish to extract information from
 #' @param offset (optional) To extract all data from the report, use 0 (default)
@@ -236,20 +282,24 @@ setGeneric("get_report", function(connection, report_id, offset=0, limit=1000) s
 #' @rdname get_report
 setMethod("get_report", "connection", function(connection, report_id, offset=0, limit=1000){
 
+  .Deprecated(new = "Report$new()",
+              msg = "This function is deprecated. The Report constructor is preferred and supports multi-table data and filtering")
+
   # Basic error checking
   if(class(report_id) != 'character') stop("'report_id' must be a character; try class(report_id)")
 
   # Get results of first report pagination
   response <- report_instance(connection=connection, report_id=report_id, offset=offset, limit=limit)
 
-  if(response$status_code != 200){
-    # Raise server error message
+  # Raises server error message
+  if(http_error(response)) {   # http != 200
     status <- http_status(response)
     errors <- content(response)
-    errmsg <- paste0("Error getting report contents.\n",
-                     "HTTP ", response$status_code, " ", status$reason, ", ", status$message, "\n",
-                     "I-Server Error ", errors$code, ", ", errors$message)
-    stop(errmsg)
+    usrmsg <- "Error getting report contents."
+
+    stop(sprintf("%s\n HTTP Error: %s %s %s\n I-Server Error: %s %s",
+                 usrmsg, response$status_code, status$reason, status$message, errors$code, errors$message),
+         call.=FALSE)
   }
 
   # report instance id, used for fetching add'l rows
@@ -284,9 +334,10 @@ setMethod("get_report", "connection", function(connection, report_id, offset=0,
 
 
 
-#' @title Extract a MicroStrategy cube into a R Data.Frame
+#' @title Extract a MicroStrategy cube into a R Data.Frame (deprecated)
 #'
-#' @description Extracts the contents of a MicroStrategy cube into a R Data.Frame
+#' @description Extracts the contents of a MicroStrategy cube into a R Data.Frame. This function is deprecated.
+#'  Check out the to_dataframe() from the Cube class. \code{\link{Cube}}
 #' @param connection MicroStrategy REST API connection object
 #' @param cube_id Unique ID of the cube you wish to extract information from
 #' @param offset (optional) To extract all data from the report, use 0 (default)
@@ -323,22 +374,27 @@ setGeneric("get_cube", function(connection, cube_id, offset=0, limit=1000) stand
 #' @rdname get_cube
 setMethod("get_cube", "connection", function(connection, cube_id, offset=0, limit=1000){
 
+  .Deprecated(new = "Cube$new()",
+              msg = "This function is deprecated. The Cube constructor is preferred and supports multi-table data and filtering")
+
   # Basic error checking
   if(class(cube_id) != 'character') stop("'cube_id' must be a character; try class(cube_id)")
 
   # Get first cube instance
   response <- cube_instance(connection=connection, cube_id=cube_id, offset=offset, limit=limit)
 
-  if(response$status_code != 200){
-    # Raise server error message
+  # Raises server error message
+  if(http_error(response)) {   # http != 200
     status <- http_status(response)
     errors <- content(response)
-    errmsg <- paste0("Error getting cube contents.\n",
-                     "HTTP ", response$status_code, " ", status$reason, ", ", status$message, "\n",
-                     "I-Server Error ", errors$code, ", ", errors$message)
-    stop(errmsg)
+    usrmsg <- "Error getting cube contents."
+
+    stop(sprintf("%s\n HTTP Error: %s %s %s\n I-Server Error: %s %s",
+                 usrmsg, response$status_code, status$reason, status$message, errors$code, errors$message),
+         call.=FALSE)
   }
 
+
   # Gets the pagination totals from the response object
   response <- content(response)
   pagination <- response$result$data$paging
@@ -371,9 +427,11 @@ setMethod("get_cube", "connection", function(connection, cube_id, offset=0, limi
 
 
 
-#' @title Create an in-memory MicroStrategy dataset
+#' @title Create an in-memory MicroStrategy dataset (deprecated)
 #'
-#' @description Creates an in-memory dataset from an R Data.Frame.
+#' @description Creates an in-memory dataset from an R Data.Frame. This function is deprecated.
+#'  Check out the add_table() & create() method from the Dataset class, which allows for uploading multi-table datasets.
+#'  \code{\link{Dataset}}
 #' @param connection MicroStrategy REST API connection object
 #' @param data_frame R Data.Frame from which an in-memory dataset will be created
 #' @param dataset_name Name of the in-memory dataset
@@ -385,6 +443,8 @@ setMethod("get_cube", "connection", function(connection, cube_id, offset=0, limi
 #'                  corresponding column name as \code{to_metric=c('myStringIntegers')}
 #' @param to_attribute (optional) Logical opposite of \code{to_metric}. Helpful for formatting an integer-based row identifier as
 #'                     a primary key in the dataset
+#' @param folder_id (optional) ID of the shared folder that the dataset should be created within. If \code{NULL}, defaults to the user's My Reports folder.
+#' @param description (optional) Description of the dataset. Must be less than or equal to 250 characters.
 #' @return Unique identifiers of the dataset and table within the newly created dataset. Required for \code{update_dataset()}
 #' @name create_dataset
 #' @rdname create_dataset
@@ -421,10 +481,14 @@ setMethod("get_cube", "connection", function(connection, cube_id, offset=0, limi
 #'                        to_attribute = c("Sepal_Length", "Sepal_Width"))
 #' }
 #' @export create_dataset
-setGeneric("create_dataset", function(connection, data_frame, dataset_name, table_name, to_metric=NULL, to_attribute=NULL) standardGeneric("create_dataset"))
+setGeneric("create_dataset", function(connection, data_frame, dataset_name, table_name,
+                                      to_metric=NULL, to_attribute=NULL, folder_id=NULL, description=NULL) standardGeneric("create_dataset"))
 
 #' @rdname create_dataset
-setMethod("create_dataset", "connection", function(connection, data_frame, dataset_name, table_name, to_metric=NULL, to_attribute=NULL){
+setMethod("create_dataset", "connection", function(connection, data_frame, dataset_name, table_name,
+                                                   to_metric=NULL, to_attribute=NULL, folder_id=NULL, description=NULL){
+  .Deprecated(new="Dataset$new()",
+              msg="This function is deprecated. The Dataset constructor is preferred and supports multi-table data.")
 
   # Basic error checking for input types
   if(class(data_frame) != "data.frame") stop("'data_frame' must be a valid R data.frame; try class(data_frame)")
@@ -436,50 +500,23 @@ setMethod("create_dataset", "connection", function(connection, data_frame, datas
   # Check column names for presence of non-word characters like ., $, ' ', ; as these will create issues in mstr
   if(any(grepl("\\W", names(data_frame)))) stop("Column names cannot have non-word characters. Try again after eliminating white space and special characters.")
 
-  # Generate microstrategy json template for the data frame
-  if(all(is.null(to_metric), is.null(to_attribute))){
-    jsonlist <- form_json(df=data_frame, table_name=table_name)
-  } else {
-    jsonlist <- form_json(df=data_frame, table_name=table_name, to_metric=to_metric, to_attribute=to_attribute)
-  }
+  # Create and transmit the dataset
+  ds <- Dataset$new(connection=connection, name=dataset_name, description=description)
+  ds$add_table(table_name, data_frame, update_policy="add", to_metric=to_metric, to_attribute=to_attribute)
+  ds$create(folder_id=folder_id)
 
-  # Encode the data in base64 encoded JSON string, required for API to work properly
-  data_encoded <- openssl::base64_encode(jsonlite::toJSON(data_frame, raw='base64'))
-
-  # create json string to pass to the server
-  json_body <- jsonlite::toJSON(list(name = dataset_name,
-                                     tables = list(list(name = table_name,
-                                                        columnHeaders = jsonlist$columnHeaders,
-                                                        data = data_encoded)),
-                                     attributes = jsonlist$attributeList,
-                                     metrics = jsonlist$metricList),
-                                auto_unbox = TRUE,
-                                pretty = TRUE)
-
-  # Make the request
-  response <- .create_dataset(connection=connection, json_body=json_body)
-
-  if(response$status_code != 200){
-    # Raise server error message
-    status <- http_status(response)
-    errors <- content(response)
-    errmsg <- paste0("Error creating the dataset.\n",
-                     "HTTP ", response$status_code, " ", status$reason, ", ", status$message, "\n",
-                     "I-Server Error code ", errors$code, ", ", errors$message)
-    stop(errmsg)
-  }
-
-  req <- content(response, as="parsed", type="application/json")
-  return(list(datasetID = req$datasetId,
-              name = req$name,
-              tableID = req$tables[[1]]$id))
+  return(list(datasetID = ds$dataset_id,
+              name = ds$name,
+              tableID = NULL))
 })
 
 
 
-#' @title Update a previously created dataset
+#' @title Update a previously created dataset (deprecated)
 #'
-#' @description Updates a previously created MicroStrategy dataset with an R Data.Frame.
+#' @description Updates a previously created MicroStrategy dataset with an R Data.Frame. This function is deprecated.
+#'  Check out the add_table() & update() & publish() method from the Dataset class, which allows for updating multi-table datasets.
+#'  \code{\link{Dataset}}
 #' @param connection MicroStrategy REST API connection object
 #' @param data_frame R Data.Frame to use to update an in-memory dataset
 #' @param dataset_id Identifier of the dataset to update, provided by \code{create_dataset()}
@@ -561,7 +598,8 @@ setGeneric("update_dataset", function(connection, data_frame, dataset_id, table_
 #' @rdname update_dataset
 setMethod("update_dataset", "connection", function(connection, data_frame, dataset_id, table_name, update_policy){
 
-  # TODO: Implement table_id or table_name
+  .Deprecated(new="Dataset$new()",
+              msg="This function is deprecated. The Dataset constructor is preferred and supports multi-table data.")
 
   # Basic error checking for input types
   if(class(data_frame) != "data.frame") stop("'data_frame' must be a valid R data.frame; try class(data_frame)")
@@ -572,30 +610,17 @@ setMethod("update_dataset", "connection", function(connection, data_frame, datas
   # Check column names for presence of non-word characters like ., $, ' ', ; as these will create issues in mstr
   if(any(grepl("\\W", names(data_frame)))) stop("Column names cannot have non-word characters. Try again after eliminating white space and special characters.")
 
-  # encode data
-  data_encoded <- openssl::base64_encode(jsonlite::toJSON(data_frame, raw='base64'))
-
-  # Generate microstrategy json template for the data frame
-  jsonlist <- form_json(df=data_frame, table_name=table_name)
-
-  # form json body
-  json_body <- jsonlite::toJSON(list(name = table_name,
-                                     columnHeaders = jsonlist$columnHeaders,
-                                     data = data_encoded),
-                                auto_unbox = TRUE,
-                                pretty = TRUE)
-
-  # make request
-  response <- .update_dataset(connection=connection, dataset_id=dataset_id, table_name=table_name,
-                              update_policy=update_policy, json_body=json_body)
-
-  if(response$status_code != 200){
-    # Raise server error message
-    status <- http_status(response)
-    errors <- content(response)
-    errmsg <- paste0("Error updating the dataset. Check for datatype inconsistencies between the data frame and the MicroStrategy dataset.\n",
-                     "HTTP ", response$status_code, " ", status$reason, ", ", status$message, "\n",
-                     "I-Server Error ", errors$code, ", ", errors$message)
-    stop(errmsg)
+  ds <- Dataset$new(connection=connection, dataset_id=dataset_id)
+  ds$add_table(name=table_name, data_frame=data_frame, update_policy=update_policy)
+  ds$update()
+  ds$publish()
+  
+  status <- 6
+  while(status != 1){
+    pub <- ds$publish_status()
+    status <- pub$status
+    if(status == 1){
+      break
+    }
   }
 })
